import asyncio
import logging
import sys
from urllib.parse import urlparse

from panoramisk import Manager
import signal

from ..handlers import EventHandler


class AmiRunner(object):
    """
    A Runner which reads Asterisk AMI events and passes them to a
    EventHandler instance.
    """
    def __init__(self, asterisk_uris, reporter, handler=EventHandler,
                 logger=None):
        """
        Create and initialize a new AmiRunner.

        Args:
            asterisk_uris (list): A list of URIs with AMI host information.
            reporter (Reporter): A reporter which should receive notifications.
            handler (class): The EventHandler class for this runner.
            logger (Logger): The log handler for the runner.
        """
        self.asterisks = asterisk_uris
        self.reporter = reporter
        self.event_handler = handler
        self.loop = asyncio.get_event_loop()
        self.logger = logger or logging.getLogger(__name__)

        self.ami_managers = {}

    def attach_all(self):
        """
        Connect to all Asterisks
        """
        for asterisk in self.asterisks:
            self.attach(asterisk)

    def attach(self, asterisk):
        """
        Set up a connection to the specified Asterisk

        Args:
            asterisk (str): A connection string
        """
        ami_host = urlparse(asterisk)

        # Create Panoramisk asterisk AMI manager.
        manager = Manager(
            loop=self.loop, host=ami_host.hostname, port=ami_host.port,
            username=ami_host.username, secret=ami_host.password,
            ssl=ami_host.scheme in ('ssl', 'tls'), encoding='utf8',
            log=self.logger)

        # Create our own channel manager.
        event_handler = self.event_handler(
            reporter=self.reporter,
            hostname=ami_host.hostname,
            logger=self.logger,
        )

        # Tell Panoramisk to which events we want to listen.
        if event_handler.FILTER_EVENTS:
            for event_name in event_handler.event_handlers().keys():
                manager.register_event(event_name, self.on_event)
        else:
            manager.register_event('*', self.on_event)

        # Record them for later use.
        self.ami_managers[manager] = event_handler

        # Tell asyncio what to work on.
        asyncio.ensure_future(manager.connect())

    def on_event(self, manager, event):
        """
        When an event comes in, pass it to the relevant channel manager.

        Args:
            manager (Manager): The AMI manager from Panoramisk.
            event (Event): AMI event (a dict-like object with event data).
        """
        assert manager in self.ami_managers
        self.ami_managers[manager].on_event(event)

    def run(self):
        """
        Start the runner and run until halted.
        """
        self.attach_all()

        signal.signal(signal.SIGINT, self._close)

        self.loop.run_forever()

    def _close(self, signal, frame):
        """
        Clean shutdown the runner.
        """
        self.logger.info('Shutting down Cacofonisk...')
        for manager in self.ami_managers:
            manager.close()
        self.reporter.close()
        sys.exit(0)



import logging

from .bridge import Bridge, BridgeDict, MissingBridgeUniqueid
from .channel import Channel, ChannelDict, MissingUniqueid
from .constants import (AST_CAUSE_ANSWERED_ELSEWHERE, AST_CAUSE_CALL_REJECTED,
                        AST_CAUSE_NORMAL_CLEARING, AST_CAUSE_NO_ANSWER,
                        AST_CAUSE_NO_USER_RESPONSE, AST_CAUSE_UNKNOWN,
                        AST_CAUSE_USER_BUSY, AST_STATE_DIALING, AST_STATE_DOWN,
                        AST_STATE_RING, AST_STATE_RINGING, AST_STATE_UP)


class EventHandler(object):
    """
    The EventHandler translates AMI events to high level call events.

    There are two ways to use EventHandler.

    The simple way to use this method is to pass your own Reporter to it.
    The Reporter receives simple, high level hooks about a call. See
    BaseReporter for more information about these hooks.

    Alternatively, you can implement your own, more low-level events by
    extending EventHandler and overriding one or more of the following methods:
    - on_state_change
    - on_a_dial
    - on_b_dial
    - on_bridge_enter
    - on_attended_transfer
    - on_blind_transfer
    - on_blonde_transfer
    - on_user_event
    - on_hangup

    See the docs about the particular methods for more information about
    what they receive and how they work.
    """
    FILTER_EVENTS = True

    def __init__(self, reporter, hostname='localhost', logger=None):
        """
        Create a EventHandler instance.

        Args:
            reporter (Reporter): A reporter to send events to.
        """
        self._reporter = reporter
        self._channels = ChannelDict()
        self._bridges = BridgeDict()
        self._logger = logger or logging.getLogger(__name__)
        self._hostname = hostname

    @classmethod
    def event_handlers(cls):
        """
        Get the events we would like to receive, and functions to handle them.

        Returns:
            dict: A dict with event names as keys and functions as values.
        """
        return {
            'FullyBooted': cls._on_fully_booted,
            # Events related to call setup.
            'Newchannel': cls._on_new_channel,
            'Newstate': cls._on_new_state,
            'LocalBridge': cls._on_local_bridge,
            'Hangup': cls._on_hangup,
            'DialBegin': cls._on_dial_begin,
            'DialEnd': cls._on_dial_end,
            # Events which change channel vars.
            'NewCallerid': cls._on_new_callerid,
            'NewAccountCode': cls._on_new_accountcode,
            'NewConnectedLine': cls._on_new_connected_line,
            # Transfer events.
            'AttendedTransfer': cls._on_attended_transfer,
            'BlindTransfer': cls._on_blind_transfer,
            # Bridges and their contents.
            'BridgeCreate': cls._on_bridge_create,
            'BridgeEnter': cls._on_bridge_enter,
            'BridgeLeave': cls._on_bridge_leave,
            'BridgeDestroy': cls._on_bridge_destroy,
            # User Events
            'UserEvent': cls.on_user_event,
        }

    def on_event(self, event):
        """
        Interpret an event, update local state and send notifications.

        This function wraps _on_event to catch exceptions related to missing
        channels and bridges, which can happen if the Cacofonisk connects to a
        running Asterisk.

        Args:
            event (dict): A dictionary containing an AMI event.
        """
        try:
            handlers = self.event_handlers()
            if event['Event'] in handlers and handlers[event['Event']]:
                handlers[event['Event']](self, event)

        except MissingUniqueid as e:
            # If this is after a recent FullyBooted and/or start of
            # self, it is reasonable to expect that certain events will
            # fail.
            self._logger.warning(
                'Channel with Uniqueid {} not in mem when processing event: '
                '{!r}'.format(e.args[0], event))
        except MissingBridgeUniqueid as e:
            # This too is reasonably expected.
            self._logger.warning(
                'Bridge with Uniqueid {} not in mem when processing event: '
                '{!r}'.format(e.args[0], event))

        self._reporter.on_event(event)

    def _on_fully_booted(self, event):
        """
        FullyBooted event is sent on successful connection to Asterisk.

        Args:
            event (dict): A FullyBooted event.
        """
        self._logger.info('Connection established to Asterisk on {}'.format(
            self._hostname))

        if len(self._bridges) > 0:
            self._logger.warning('Bridge buffers not empty! Flushing {} '
                                 'bridges.'.format(len(self._bridges)))
            self._bridges = BridgeDict()

        if len(self._channels) > 0:
            self._logger.warning('Channel buffers not empty! Flushing {} '
                                 'channels.'.format(len(self._channels)))
            self._channels = ChannelDict()

    def _on_new_channel(self, event):
        """
        NewChannel event is sent when Asterisk creates a new channel.

        If this event is received, a new Channel will be created and
        populated using the data in this event.

        Args:
            event (dict): A NewChannel event.
        """
        channel = Channel(event)
        self._channels[channel.uniqueid] = channel

    def _on_new_state(self, event):
        """
        NewState event is sent when the state of a channel changes.

        In the phase where a phone dials, rings and is picked up,
        the channel state changes several times.

        Args:
            event (dict): A NewState event.
        """
        channel = self._channels[event['Uniqueid']]

        old_state = channel.state
        channel.state = int(event['ChannelState'])
        assert old_state != channel.state

        self.on_state_change(channel, old_state)

    def _on_local_bridge(self, event):
        """
        A LocalBridge is sent when two local channels are bridge.

        Asterisk Local Channels always have two semi-channels, identified
        with the ;1 and ;2 suffixes. When a call is set up, a LocalBridge is
        created to link both semi-channels together.

        Args:
            event (dict): A LocalBridge event.
        """
        local_one = self._channels[event['LocalOneUniqueid']]
        local_two = self._channels[event['LocalTwoUniqueid']]

        assert local_one.fwd_local_bridge is None
        assert local_one.back_local_bridge is None
        assert local_two.fwd_local_bridge is None
        assert local_two.back_local_bridge is None

        local_one.fwd_local_bridge = local_two
        local_two.back_local_bridge = local_one

    def _on_hangup(self, event):
        """
        A Hangup event is sent when a channel is hung up.

        Args:
            event (dict): A Hangup event.
        """
        channel = self._channels[event['Uniqueid']]

        self.on_hangup(channel, event)

        # Disconnect all channels linked to this channel.
        if channel.fwd_local_bridge:
            channel.fwd_local_bridge.back_local_bridge = None

        if channel.back_local_bridge:
            channel.back_local_bridge.fwd_local_bridge = None

        # Remove the channel from our own list.
        del self._channels[channel.uniqueid]

        # If we don't have any channels, check if we're completely clean.
        if not len(self._channels):
            self._logger.info('(no channels left)')

    def _on_dial_begin(self, event):
        """
        A DialBegin event is sent when a dial is created between two channels.

        Dials are created to connect channels during the call setup phase.
        In a simple call scenario, you'd see two dials being created: a dial
        from the calling SIP channel to Local channel semi one, and from
        Local channel semi two to the target SIP channel. The Local semi
        channels are connected using a LocalBridge. By tracing the dials and
        local bridges, it's possible to follow the relation between a
        calling channel and it's targets.

        Args:
            event (dict): A DialBegin event.
        """
        if 'DestUniqueid' in event:
            if 'Uniqueid' in event:
                # This is a dial between two channels. So let's link them
                # together.
                channel = self._channels[event['Uniqueid']]
                destination = self._channels[event['DestUniqueid']]
                channel.is_calling = True

                # Verify target is not being dialed already.
                assert not destination.back_dial

                # _fwd_dials is a list of channels being dialed by A.
                channel.fwd_dials.append(destination)

                # _back_dial is the channel dialing B.
                destination.back_dial = channel

                self.on_dial_begin(channel, destination)
            else:
                # The dial has a destination but not source. That means this
                # Dial was created by an Originate.
                destination = self._channels[event['DestUniqueid']]
                destination.is_originated = True
        else:
            raise AssertionError(
                'A DialBegin event was generated without DestUniqueid: '
                '{}'.format(event))

    def _on_dial_end(self, event):
        """
        DialEnd event is sent when a dial is disconnected between two channels.

        When a call is answered (picked up) or hung up (without being picked
        up), the dial between two channels is ended. After a DialEnd, you will
        either see multiple Hangup events, or see bridges being set up to
        connect the Channels.

        Args:
            event (dict): A DialEnd event.
        """
        # Check if we have a source and destination channel to pull
        # apart. Originate creates Dials without source.
        if 'Uniqueid' in event and 'DestUniqueid' in event:
            channel = self._channels[event['Uniqueid']]
            destination = self._channels[event['DestUniqueid']]

            destination.back_dial = None

            if destination in channel.fwd_dials:
                channel.fwd_dials.remove(destination)
        else:
            # Dials without Uniqueid and DestUniqueid can occur, but we
            # can't/don't handle them.
            pass

    def _on_attended_transfer(self, event):
        """
        An AttendedTransfer event is sent after attended and blonde transfers.

        Args:
            event (dict): An AttendedTransfer event.
        """
        orig_transferer = self._channels[event['OrigTransfererUniqueid']]
        second_transferer = self._channels[
            event['SecondTransfererUniqueid']]

        if event['DestType'] == 'Bridge':
            self.on_attended_transfer(
                orig_transferer, second_transferer, event)
        elif event['DestType'] == 'App' and event['DestApp'] == 'Dial':
            self.on_blonde_transfer(
                orig_transferer, second_transferer, event)
        else:
            raise NotImplementedError(event)

    def _on_blind_transfer(self, event):
        """
        A BlindTransfer event is sent after blind transfers.

        Args:
            event (dict): A BlindTransfer event.
        """
        transferer = self._channels[event['TransfererUniqueid']]
        transferee = self._channels[event['TransfereeUniqueid']]

        self.on_blind_transfer(transferer, transferee, event)

    def _on_bridge_create(self, event):
        """
        A BridgeCreate event is sent when Asterisk creates a new bridge.

        Args:
            event (dict): A BridgeCreate event.
        """
        assert event['BridgeUniqueid'] not in self._bridges
        bridge = Bridge(event)
        self._bridges[bridge.uniqueid] = bridge

    def _on_bridge_enter(self, event):
        """
        A BridgeEnter event is sent when a channel joins a bridge.

        Args:
            event (dict): A BridgeEnter event.
        """
        channel = self._channels[event['Uniqueid']]
        bridge = self._bridges[event['BridgeUniqueid']]

        bridge.peers.add(channel)
        channel.bridge = bridge

        self.on_bridge_enter(channel, bridge)

    def _on_bridge_leave(self, event):
        """
        A BridgeLeave event is sent when a channel is removed from a bridge.

        Args:
            event (dict): A BridgeLeave event.
        """
        channel = self._channels[event['Uniqueid']]
        bridge = self._bridges[event['BridgeUniqueid']]

        bridge.peers.remove(channel)
        channel.bridge = None

    def _on_bridge_destroy(self, event):
        """
        A BridgeDestroy event is sent when a bridge is removed by Asterisk.

        Args:
            event (dict): A BridgeDestroy event.
        """
        assert len(self._bridges[event['BridgeUniqueid']]) is 0
        del self._bridges[event['BridgeUniqueid']]

    def _on_new_callerid(self, event):
        """
        A NewCallerid event is sent when the CallerID of a channel changes.

        Args:
            event (dict): A NewCallerid event.
        """
        channel = self._channels[event['Uniqueid']]

        channel.caller_id = channel.caller_id.replace(
            name=event['CallerIDName'],
            num=event['CallerIDNum'],
        )
        channel.cid_calling_pres = event['CID-CallingPres']

    def _on_new_connected_line(self, event):
        """
        A NewConnectedLine event is sent when the ConnectedLine changes.

        Args:
            event (dict): A NewConnectedLine event.
        """
        channel = self._channels[event['Uniqueid']]

        channel.connected_line = channel.connected_line.replace(
            name=event['ConnectedLineName'],
            num=event['ConnectedLineNum'],
        )

    def _on_new_accountcode(self, event):
        """
        A NewAccountCode is sent when the AccountCode of a channel changes.

        Args:
            event (dict): A NewAccountCode event.
        """
        channel = self._channels[event['Uniqueid']]

        channel.account_code = event['AccountCode']

    # ===================================================================
    # Actual event handlers you can override
    # ===================================================================

    def on_state_change(self, channel, old_state):
        """
        Handle the change of a ChannelState.

        If the status goes from DOWN to RING or UP, then it means the calling
        party hears a dialing tone. If the status goes from DOWN to RINGING
        or UP, then it means the phone of a called party is starting to ring.

        Args:
            channel (Channel): The channel being changed.
            old_state (int): The state before the state change.
        """
        if channel.is_local:
            return

        new_state = channel.state

        if old_state == AST_STATE_DOWN:
            if new_state in (AST_STATE_DIALING, AST_STATE_RING, AST_STATE_UP):
                self.on_a_dial(channel)
            elif new_state == AST_STATE_RINGING:
                self.on_b_dial(channel)

    def on_a_dial(self, channel):
        """
        Handle the event where the caller phone hears the ring tone.

        We don't want this. It's work to get all the values right, and
        when we do, this would look just like on_b_dial.

        Args:
            channel (Channel):
        """
        pass

    def on_b_dial(self, channel):
        """
        Handle the event where a callee phone starts to ring.

        In our case, we check if a dial has already been set up for the
        channel. If so, we may want to send a ringing event.

        Args:
            channel (Channel): The channel of the B side.
        """
        if channel.back_dial:
            self.on_b_dial_ringing(channel)

    def on_dial_begin(self, channel, destination):
        """
        Handle an event where a dial is set up.

        In our case, we check if the channel already has state ringing.
        If so, we may want to send a ringing event.

        Args:
            channel (Channel): The channel initiating the dial.
            destination (Channel): The channel being dialed.
        """
        if not destination.is_local and destination.state == 5:
            self.on_b_dial_ringing(destination)

    def on_b_dial_ringing(self, channel):
        """
        Check a ringing channel and sent a ringing event if required.

        By default, this function will only be called if the destination
        channel:

        - Has an open dial (so a way to trace back how it's being called).
        - Has state "ringing".

        Args:
            channel (Channel): The channel being dialed.
        """
        if 'ignore_b_dial' in channel.custom:
            # Notifications were already sent for this channel.
            # Unset the flag and move on.
            del (channel.custom['ignore_b_dial'])
            return

        a_chan = channel.get_dialing_channel()

        if 'raw_blind_transfer' in a_chan.custom:
            # This is an interesting exception: we got a Blind Transfer
            # message earlier and recorded it in this attribute. We'll
            # translate this b_dial to first a on_b_dial and then the
            # on_transfer event.
            transferer = a_chan.custom.pop('raw_blind_transfer')

            target_chans = a_chan.get_dialed_channels()

            for target in target_chans:
                # To prevent notifications from being sent multiple times,
                # we set a flag on all other channels except for the one
                # starting to ring right now.
                if target != channel:
                    target.custom['ignore_b_dial'] = True

            self._reporter.on_blind_transfer(
                caller=a_chan.as_namedtuple(),
                transferer=transferer.as_namedtuple(),
                targets=[chan.as_namedtuple() for chan in target_chans],
            )
        elif (
                a_chan.is_originated and
                a_chan.fwd_dials and a_chan.fwd_local_bridge
        ):
            # Calls setup through Originate are harder to track.
            # The Channel passed to the Originate has two semis. The Context
            # channel is called first, and when it's up, put in a bridge
            # with semi 2. Next, semi 1 will dial out to the other party.

            # To make it look like a normal call, we will show the call from
            # the Context as the calling party and the call from Channel as
            # the called party.
            originating_chan = a_chan
            a_bridge = originating_chan.fwd_local_bridge.bridge
            a_chan = [peer for peer in a_bridge.peers
                      if not peer.is_local][0]

            if not a_chan.is_local:
                called_exten = originating_chan.fwd_dials[0].exten
                a_chan.exten = called_exten
                a_chan.is_calling = True

                if not a_chan.has_extension:
                    self._logger.error(
                        'Caller (Originate) did not have an extension: '
                        '{}'.format(channel))

                self._reporter.on_b_dial(
                    caller=a_chan.as_namedtuple(),
                    targets=[channel.as_namedtuple()],
                )
        elif not a_chan.is_local:
            # We'll want to send one ringing event for all targets, so send
            # one notification and mark the rest as already notified.
            open_dials = a_chan.get_dialed_channels()
            targets = [dial.as_namedtuple() for dial in open_dials]

            if not a_chan.has_extension:
                self._logger.error(
                    'Caller (Dial) did not have an extension: {}'.format({
                        'caller': a_chan.as_namedtuple(),
                        'destination': channel.as_namedtuple(),
                    }))

            if not targets:
                self._logger.error(
                    'Caller (Dial) did not have any dialed channels: '
                    '{}'.format({
                        'caller': a_chan.as_namedtuple(),
                        'destination': channel.as_namedtuple(),
                    }))

            self._reporter.on_b_dial(
                caller=a_chan.as_namedtuple(),
                targets=targets,
            )

            for b_chan in open_dials:
                if b_chan != channel:
                    # To prevent notifications from being sent multiple
                    # times, we set a flag on all other channels except
                    # for the one starting to ring right now.
                    b_chan.custom['ignore_b_dial'] = True

    def on_bridge_enter(self, channel, bridge):
        """
        Post-process a BridgeEnter event to notify of a call in progress.

        This function will check if the bridge already contains other SIP
        channels. If so, it's interpreted as a call between two channels
        being connected.

        WARNING: This function does not behave as desired for
        bridges with 3+ parties, e.g. conference calls.

        Args:
            channel (Channel): The channel entering the bridge.
            bridge (Bridge): The bridge the channel enters.
        """
        sip_peers = channel.get_bridge_peers_recursive()

        if len(sip_peers) < 2:
            # There are not enough interesting channels to form a call.
            return

        callers = set([peer for peer in sip_peers if peer.is_calling])
        targets = sip_peers - callers

        if len(callers) > 1:
            # Hmm, we have multiple callers. This can happen on an
            # AB-CB-AC transfer. Let's do something ugly.

            # Our oldest caller is going to be the new caller.
            sorted_callers = sorted(
                callers, key=lambda chan: chan.name.rsplit('-', 1)[1])
            caller = sorted_callers.pop(0)

            # The rest are will be marked as targets.
            for non_caller in sorted_callers:
                targets.add(non_caller)
        elif len(callers) < 1:
            # A call should always have a caller.
            self._logger.warning('Call {} has too few callers: {}'.format(
                channel.linkedid, len(callers)))
            return
        else:
            caller = next(iter(callers))

        if len(targets) != 1:
            # This can happen with a conference call, but is not supported.
            self._logger.warning('Call {} has {} targets.'.format(
                channel.linkedid, len(targets)))
            return
        else:
            target = next(iter(targets))

        # Check and set a flag to prevent the event from being fired again.
        if 'is_picked_up' not in caller.custom:
            caller.custom['is_picked_up'] = True

            self._reporter.on_up(
                caller=caller.as_namedtuple(),
                target=target.as_namedtuple(),
            )

    def on_attended_transfer(self, orig_transferer, second_transferer, event):
        """
        Gets invoked when an attended transfer is completed.

        In an attended transfer, one of the participants of a conversation
        calls a third participant, waits for the third party to answer, talks
        to the third party and then transfers their original conversation
        partner to the third party.

        Args:
            orig_transferer (Channel): The original channel is the channel
            which the redirector used to talk with the person who's being
            transferred.
            second_transferer (Channel): The target channel is the channel
            which the redirector used to set up the call to the person to
            whom the call is being transferred.
            event (dict): The data of the AttendedTransfer event.
        """
        if 'TransfereeUniqueid' in event and 'TransferTargetUniqueid' in event:
            # Nice, Asterisk just told us who the transferee and transfer
            # target are. Let's just do what Asterisk says.
            transferee = self._channels[event['TransfereeUniqueid']]
            target = self._channels[event['TransferTargetUniqueid']]
        else:
            # Ouch, Asterisk didn't tell us who is the transferee and who is
            #  the target, which means we need to figure it out ourselves.

            # We can find both channels in the Destination Bridge.
            target_bridge = self._bridges[event['DestBridgeUniqueid']]

            if len(target_bridge) < 2:
                self._logger.warning(
                    'Attn Xfer DestBridge does not have enough peers for '
                    'event: {!r}'.format(event))
                return

            peer_one, peer_two = target_bridge.peers

            # The next challenge now is to figure out which channel is the
            # transferee and which one is the target..
            if peer_one.linkedid == event['OrigTransfererLinkedid']:
                # Peer one has the same linkedid as the call before the
                # transfer, so it must be the transferee.
                transferee = peer_one
                target = peer_two
            elif peer_two.linkedid == event['OrigTransfererLinkedid']:
                transferee = peer_two
                target = peer_one
            else:
                raise NotImplementedError(
                    'Could not determine caller and target after attended '
                    'transfer - OrigTransfererLinkedid not found. '
                    'OrigTransferer: {}. SecondTransferer: {}. Peers: {}. '
                    'Event: {}.'.format(
                        orig_transferer, second_transferer,
                        target_bridge.peers, event,
                    )
                )

        if transferee.is_local:
            # If the channels were not optimized before the transfer was
            # started, the transferee may be a local channel. To generate the
            # right event, we need to find the real SIP channel which started
            # the transfer.
            for peer in transferee.get_bridge_peers_recursive():
                if peer not in (target, second_transferer, orig_transferer):
                    transferee = peer

        transferee.is_calling = True
        # transferee becomes the new caller, so it should have a valid
        # extension. We can use set it to one of the transfer extensions.
        if event['SecondTransfererExten']:
            transferee.exten = event['SecondTransfererExten']
        elif event['OrigTransfererExten']:
            transferee.exten = event['OrigTransfererExten']
        else:
            transferee.exten = event['TransferTargetCallerIDNum']

        if not transferee.has_extension:
            self._logger.error(
                'Transferee (attn xfer) did not have an extension: '
                '{}'.format(transferee))

        # In some transfer scenarios, a caller can become a target. Because
        # of that, we need to make sure the target is not marked as calling.
        target.is_calling = False

        self._reporter.on_attended_transfer(
            caller=transferee.as_namedtuple(),
            transferer=second_transferer.as_namedtuple(),
            target=target.as_namedtuple(),
        )

        # Prevent a hangup event from being fired for the transfer channels.
        orig_transferer.custom['ignore_a_hangup'] = True
        second_transferer.custom['ignore_a_hangup'] = True

    def on_blind_transfer(self, transferer, transferee, event):
        """
        Handle a blind (cold) transfer event.

        In a blind transfer, one of the call participant transfers their
        conversation partner to a third party. However, unlike with an
        attended transfer, the redirector doesn't wait for the other end to
        pick up, but just punches in the number and sends their conversation
        party away. Because of this, multiple phones may actually be addressed
        by this transfer, hence the multiple targets. The real participant can
        be recovered later on when someone answers the transferred call.

        This Transfer event is earlier than the dial. We mark it and
        wait for the b_dial event. In on_b_dial we send out both the
        on_b_dial and the on_transfer.

        Args:
            transferer (Channel): The channel referring to another extension.
            transferee (Channel): The channel being referred.
            event (dict): The data of the BlindTransfer event.
        """
        transferee.custom['raw_blind_transfer'] = transferer

        # Remove the is_picked_up flag so we can figure a new in-progress
        # event when the transfer target picks up.
        try:
            del transferee.custom['is_picked_up']
        except KeyError:
            pass

        # Prevent a hangup event from being fired for the transfer channels.
        transferer.custom['ignore_a_hangup'] = True

        # Make it look like the transferee is calling the transfer extension.
        transferee.is_calling = True

        transferee.exten = event['Extension']
        transferer.exten = event['Extension']

    def on_blonde_transfer(self, orig_transferer, second_transferer, event):
        """
        Handle the AttendedTransfer event for a blonde transfer.

        A blonde is a middle road between blind transfers and attended
        transfers. With a blond transfer, the redirector requests an attended
        transfer but doesn't wait for the receiving end to pick up. Since the
        data of blind and blonde transfers looks identical, they don't have
        special hooks.

        Args:
            orig_transferer (Channel): The original channel is the channel
            which the redirector used to talk with the person who's being
            transferred.
            second_transferer (Channel): The target channel is the channel
            which the redirector used to set up the call to the person to
            whom the call is being transferred.
            event (dict): The AttendedTransfer event.
        """
        transferee = self._channels[event['TransfereeUniqueid']]

        # Remove the is_picked_up flag so we can figure a new in-progress
        # event when the transfer target picks up.
        try:
            del transferee.custom['is_picked_up']
        except KeyError:
            pass

        # Make it look like the transferee is calling the transfer extension.
        transferee.is_calling = True
        transferee.exten = second_transferer.exten

        if not transferee.has_extension:
            self._logger.error(
                'Transferee (blonde xfer) did not have an extension: '
                '{}'.format(transferee))

        targets = second_transferer.get_dialed_channels().union(
            transferee.get_dialed_channels())

        self._reporter.on_blonde_transfer(
            caller=transferee.as_namedtuple(),
            transferer=second_transferer.as_namedtuple(),
            targets=[target.as_namedtuple() for target in targets],
        )

        # Prevent a hangup event from being fired for the transfer channels.
        orig_transferer.custom['ignore_a_hangup'] = True
        second_transferer.custom['ignore_a_hangup'] = True

    def on_user_event(self, event):
        """
        Handle custom UserEvent messages from Asterisk.

        Adding user events to a dial plan is a useful way to send additional
        information to Cacofonisk. You could add additional user info,
        parameters used for processing the events and more.

        Args:
            event (dict): Dict-like object with all attributes of the event.
        """
        self._reporter.on_user_event(event)

    def on_hangup(self, channel, event):
        """
        Process the channel just before it is hung up.

        Args:
            channel (Channel): The channel being disconnected.
            event (dict): The Hangup event data.
        """
        if channel.is_local:
            return

        if 'raw_blind_transfer' in channel.custom:
            # Panic! This channel had a blind transfer coming up but it's
            # being hung up! That probably means the blind transfer target
            # could not be reached.
            # Ideally, we would simulate a full blind transfer having been
            # completed but hung up with an error. However, no channel
            # to the third party has been created.
            redirector = channel.custom.pop('raw_blind_transfer')

            a_chan = redirector if redirector.is_calling else channel

            # TODO: Maybe give another status code than 'completed' here?
            self._reporter.on_hangup(
                caller=a_chan.as_namedtuple(),
                reason='completed',
            )

        elif 'ignore_a_hangup' in channel.custom:
            # This is a calling channel which performed an attended
            # transfer. Because the call has already been "hung up"
            # with the transfer, we shouldn't send a hangup notification.
            pass

        elif channel.is_calling:
            # The caller is being disconnected, so we should notify the
            # user.
            self._reporter.on_hangup(
                caller=channel.as_namedtuple(),
                reason=self._hangup_reason(channel, event),
            )

    def _hangup_reason(self, channel, event):
        """
        Map the Asterisk hangup causes to easy to understand strings.

        Args:
            channel (Channel): The channel which is hung up.
            event (Event): The data of the event.
        """
        hangup_cause = int(event['Cause'])

        # See https://wiki.asterisk.org/wiki/display/AST/Hangup+Cause+Mappings
        if hangup_cause == AST_CAUSE_NORMAL_CLEARING:
            # If channel is not up, the call never really connected.
            # This happens when call confirmation is unsuccessful.
            if channel.state == AST_STATE_UP:
                return 'completed'
            else:
                return 'no-answer'
        elif hangup_cause == AST_CAUSE_USER_BUSY:
            return 'busy'
        elif hangup_cause in (AST_CAUSE_NO_USER_RESPONSE, AST_CAUSE_NO_ANSWER):
            return 'no-answer'
        elif hangup_cause == AST_CAUSE_ANSWERED_ELSEWHERE:
            return 'answered-elsewhere'
        elif hangup_cause == AST_CAUSE_CALL_REJECTED:
            return 'rejected'
        elif hangup_cause == AST_CAUSE_UNKNOWN:
            # Sometimes Asterisk doesn't set a proper hangup cause.
            # If our a_chan is already up, this probably means the
            # call was successful. If not, that means the caller hung up,
            # which we assign the "cancelled" status.
            if channel.state == AST_STATE_UP:
                return 'completed'
            else:
                return 'cancelled'
        else:
            return 'failed'



from cacofonisk import BaseReporter
from tests.replaytest import ChannelEventsTestCase


class TestVarSet(ChannelEventsTestCase):

    def test_user_event(self):
        """Test UserEvents are passed on to the reporter.

        The Asterisk event type UserEvent is reserved for custom events
        defined by the user. You can use it to pass additional data from a
        dial plan to Cacofonisk (like whether it should send events for a
        given call).
        """
        class UserEventReporter(BaseReporter):
            def __init__(self):
                super(UserEventReporter, self).__init__()
                self.events = []

            def on_user_event(self, event):
                desired = ('UserEvent', 'Provider', 'AccountCode',
                           'WebhookUrls', 'Direction', 'ClientId',
                           'AccountInternalNumber', 'UserInternalNumbers')
                event = {key: event[key] for key in desired if key in event}
                self.events.append(event)

        events = self.run_and_get_events(
            'fixtures/var_set/user_events.json',
            UserEventReporter()
        )

        expected_events = [
            {
                'AccountCode': '260010001',
                'AccountInternalNumber': '201',
                'ClientId': '26001',
                'Direction': 'outbound',
                'Provider': 'webhook',
                'UserEvent': 'NotifyCallstate',
                'WebhookUrls': '',
            },
            {
                'AccountCode': '15001',
                'ClientId': '15001',
                'Direction': 'inbound',
                'UserEvent': 'NotifyCallstate',
                'WebhookUrls': 'http://example.com/foo/bar',
            },
            {
                'AccountCode': '150010001',
                'Provider': 'cloudcti',
                'UserEvent': 'NotifyCallstate',
            },
            {
                'AccountCode': '150010001',
                'AccountInternalNumber': '201',
                'Provider': 'webhook',
                'UserEvent': 'NotifyCallstate',
                'UserInternalNumbers': '601',
            },
        ]

        self.assertEqual(expected_events, events)



from collections import namedtuple


class CallerId(namedtuple('CallerIdBase', 'name num')):
    """
    CallerId holds immutable information about one end of a call.

    The users of this application are interested in tuples of caller ID number,
    caller ID name and sometimes an account ID (accountcode), and it's privacy
    settings, both for call initiators and for call recipients.

    Usage::

        caller = CallerId(name='My name', number='+311234567', is_public=True)
        caller = caller.replace(code=123456789)
    """

    def __new__(cls, name='', num=''):
        if name == '<unknown>':
            name = ''

        if num == '<unknown>':
            num = ''

        return super().__new__(cls, name, str(num))

    def replace(self, **kwargs):
        """
        Create a copy of this CallerId with specified changes.

        Args:
            **kwargs: One or more of code, name, number, is_public.

        Returns:
            CallerId: A new instance with replaced values.
        """
        if 'name' in kwargs and kwargs['name'] == '<unknown>':
            kwargs['name'] = ''

        if 'num' in kwargs and kwargs['num'] == '<unknown>':
            kwargs['num'] = ''

        return self._replace(**kwargs)



import logging


class BaseReporter(object):
    def close(self):
        """
        Called on end, so any buffered output can be flushed.
        """
        pass

    def on_event(self, event):
        """
        Called after the regular even processing has been done.

        Useful when custom event processing is required in addition to
        regular event processing.

        Args:
            event (Message): Dict-like object with all attributes in the event.
        """
        pass

    def on_b_dial(self, caller, targets):
        """
        Gets invoked when the B side of a call is initiated.

        In the common case, calls in Asterisk consist of two sides: A calls
        Asterisk and Asterisk calls B. This event is fired when Asterisk
        performs the second step.

        Args:
            caller (SimpleChannel): The initiator of the call.
            targets (list): The recipients of the call.
        """
        pass

    def on_up(self, caller, target):
        """
        Track when a call has been set up between two parties.

        In simple calls, a "up" event is raised when a call has been ringing
        and it has been answered. Additionally, around the transfer of a call
        a new "up" event is raised as well not notify who are still calling.

        Args:
            caller (SimpleChannel): The initiator of the call.
            target (SimpleChannel): The recipient of the call.
        """
        pass

    def on_attended_transfer(self, caller, transferer, target):
        """
        Gets invoked when an attended transfer is completed.

        In an attended transfer, one of the participants of a conversation
        calls a third participant, waits for the third party to answer, talks
        to the third party and then transfers their original conversation
        partner to the third party.

        Args:
            caller (SimpleChannel): The party which has been transferred.
            transferer (SimpleChannel): The party who initiated the transfer.
            target (SimpleChannel): The party which received the transfer.
        """
        pass

    def on_blonde_transfer(self, caller, transferer, targets):
        """
        Gets invoked when a blonde transfer is completed.

        A blonde transfer is a transfer which is started like an attended
        transfer, but the transferer steps out of the call before it is
        picked up.

        Args:
            caller (SimpleChannel): The party being transferred.
            transferer (SimpleChannel): The party initiating the transfer.
            targets (list): The channels being dialed.
        """
        pass

    def on_blind_transfer(self, caller, transferer, targets):
        """
        Gets invoked when a blind transfer is completed.

        A blind transfer is transfer where the transferer tells Asterisk to
        refer the call to another number and immediately ends the call. No
        b_dial is triggered before the transfer is completed.

        Args:
            caller (SimpleChannel): The party being transferred.
            transferer (SimpleChannel): The party initiating the transfer.
            targets (list): The channels being dialed.
        """
        pass

    def on_user_event(self, event):
        """
        Handle custom UserEvent messages from Asterisk.

        Adding user events to a dial plan is a useful way to send additional
        information to Cacofonisk. You could add additional user info,
        parameters used for processing the events and more.

        Args:
            event (Message): Dict-like object with all attributes in the event.
        """
        pass

    def on_hangup(self, caller, reason):
        """
        Track when a call ends for a caller.

        Args:
            caller (SimpleChannel): The initiator of the call.
            reason (str): A textual reason as to why the call was ended.
        """
        pass


class LoggingReporter(BaseReporter):
    """
    LoggingReporter is a simple base reporter which logs all calls to the
    provided Logger instance.
    """
    def __init__(self, logger=None):
        """
        Create a logger for this reporter.

        Args:
            logger (Logger): An optional logger instance.
        """
        self._logger = logger or logging.getLogger(__name__)

    def close(self):
        """
        Called on end, so any buffered output can be flushed.
        """
        pass

    def on_event(self, event):
        """
        Called after the regular even processing has been done.

        Useful when custom event processing is required in addition to
        regular event processing.

        Args:
            event (Message): Dict-like object with all attributes in the event.
        """
        pass

    def on_b_dial(self, caller, targets):
        """
        Gets invoked when the B side of a call is initiated.

        In the common case, calls in Asterisk consist of two sides: A calls
        Asterisk and Asterisk calls B. This event is fired when Asterisk
        performs the second step.

        Args:
            caller (SimpleChannel): The initiator of the call.
            targets (list): The recipients of the call.
        """
        self._logger.info('{} ringing: {} --> {} ({})'.format(
            caller.linkedid, caller, caller.exten, targets))

    def on_up(self, caller, target):
        """
        Track when a call has been set up between two parties.

        In simple calls, a "up" event is raised when a call has been ringing
        and it has been answered. Additionally, around the transfer of a call
        a new "up" event is raised as well not notify who are still calling.

        Args:
            caller (SimpleChannel): The initiator of the call.
            target (SimpleChannel): The recipient of the call.
        """
        self._logger.info('{} up: {} --> {} ({})'.format(
            caller.linkedid, caller, caller.exten, target))

    def on_attended_transfer(self, caller, transferer, target):
        """
        Gets invoked when an attended transfer is completed.

        In an attended transfer, one of the participants of a conversation
        calls a third participant, waits for the third party to answer, talks
        to the third party and then transfers their original conversation
        partner to the third party.

        Args:
            caller (SimpleChannel): The party which has been transferred.
            transferer (SimpleChannel): The party who initiated the transfer.
            target (SimpleChannel): The party which received the transfer.
        """
        self._logger.info(
            '{} <== {} attn xfer: {} <--> {} (through {})'.format(
                caller.linkedid, transferer.linkedid,
                caller, target, transferer))

    def on_blonde_transfer(self, caller, transferer, targets):
        """
        Gets invoked when a blonde transfer is completed.

        A blonde transfer is a transfer which is started like an attended
        transfer, but the transferer steps out of the call before it is
        picked up.

        Args:
            caller (SimpleChannel): The party being transferred.
            transferer (SimpleChannel): The party initiating the transfer.
            targets (list): The channels being dialed.
        """
        self._logger.info(
            '{} <== {} blonde xfer: {} <--> {} (through {})'.format(
                caller.linkedid, transferer.linkedid, caller, targets,
                transferer))

    def on_blind_transfer(self, caller, transferer, targets):
        """
        Gets invoked when a blind transfer is completed.

        A blind transfer is transfer where the transferer tells Asterisk to
        refer the call to another number and immediately ends the call. No
        b_dial is triggered before the transfer is completed.

        Args:
            caller (SimpleChannel): The party being transferred.
            transferer (SimpleChannel): The party initiating the transfer.
            targets (list): The channels being dialed.
        """
        self._logger.info('{} bld xfer: {} <--> {} (through {})'.format(
            caller.linkedid, caller, targets, transferer))

    def on_user_event(self, event):
        """
        Handle custom UserEvent messages from Asterisk.

        Adding user events to a dial plan is a useful way to send additional
        information to Cacofonisk. You could add additional user info,
        parameters used for processing the events and more.

        Args:
            event (Message): Dict-like object with all attributes in the event.
        """
        self._logger.info('{} user_event: {}'.format(event['Linkedid'], event))

    def on_hangup(self, caller, reason):
        """
        Track when a call ends for a caller.

        Args:
            caller (SimpleChannel): The initiator of the call.
            reason (str): A textual reason as to why the call was ended.
        """
        self._logger.info('{} hangup: {} --> {} (reason: {})'.format(
            caller.linkedid, caller, caller.exten, reason))


class MultiReporter(LoggingReporter):
    """
    MultiReporter is a reporter which combines multiple reporters and
    forwards received events to all of them.
    """
    def __init__(self, reporters, logger=None):
        """
        Create a multi reporter with the given reporters.

        Args:
            reporters (list): An iterable with reporters.
            logger (Logger): Not used.
        """
        super(MultiReporter, self).__init__(logger=logger)

        self.reporters = reporters

    def close(self):
        super(MultiReporter, self).close()

        for reporter in self.reporters:
            reporter.close()

    def on_event(self, event):
        super(MultiReporter, self).on_event(event)

        for reporter in self.reporters:
            reporter.on_event(event)

    def on_user_event(self, event):
        super(MultiReporter, self).on_user_event(event)

        for reporter in self.reporters:
            reporter.on_user_event(event)

    def on_b_dial(self, caller, targets):
        super(MultiReporter, self).on_b_dial(caller, targets)

        for reporter in self.reporters:
            reporter.on_b_dial(caller, targets)

    def on_up(self, caller, target):
        super(MultiReporter, self).on_up(caller, target)

        for reporter in self.reporters:
            reporter.on_up(caller, target)

    def on_attended_transfer(self, caller, transferer, target):
        super(MultiReporter, self).on_attended_transfer(
            caller, transferer, target)

        for reporter in self.reporters:
            reporter.on_attended_transfer(caller, transferer, target)

    def on_blonde_transfer(self, caller, transferer, targets):
        super(MultiReporter, self).on_blonde_transfer(
            caller, transferer, targets)

        for reporter in self.reporters:
            reporter.on_blonde_transfer(caller, transferer, targets)

    def on_blind_transfer(self, caller, transferer, targets):
        super(MultiReporter, self).on_blind_transfer(
            caller, transferer, targets)

        for reporter in self.reporters:
            reporter.on_blind_transfer(caller, transferer, targets)

    def on_hangup(self, caller, reason):
        super(MultiReporter, self).on_hangup(caller, reason)

        for reporter in self.reporters:
            reporter.on_hangup(caller, reason)



from cacofonisk.callerid import CallerId
from cacofonisk.channel import SimpleChannel
from tests.replaytest import ChannelEventsTestCase


class TestMobile(ChannelEventsTestCase):

    def test_simple_mobile(self):
        """
        Test a call from external which is routed to another number.
        """
        events_file = 'fixtures/mobile/simple_mobile.json'
        events = self.run_and_get_events(events_file)

        caller = SimpleChannel(
            name='SIP/voipgrid-siproute-docker-00000036',
            uniqueid='f29ea68048f6-1530017224.892',
            linkedid='f29ea68048f6-1530017224.892',
            account_code='15001',
            caller_id=CallerId(num='+31150010001'),
            cid_calling_pres=None,
            connected_line=CallerId(),
            exten='+31150010004',
            state=6,
        )

        target = SimpleChannel(
            name='SIP/voipgrid-siproute-docker-00000037',
            uniqueid='f29ea68048f6-1530017224.905',
            linkedid='f29ea68048f6-1530017224.892',
            account_code='15001',
            caller_id=CallerId(num='+31260010001'),
            cid_calling_pres='0 (Presentation Allowed, Not Screened)',
            connected_line=CallerId(num='+31150010001'),
            exten='s',
            state=6,
        )

        expected_events = [
            ('on_b_dial', {
                'caller': caller.replace(state=4),
                'targets': [target.replace(state=5)],
            }),
            ('on_up', {
                'caller': caller,
                'target': target,
            }),
            ('on_hangup', {
                'caller': caller,
                'reason': 'completed',
            }),
        ]

        self.assertEqual(expected_events, events)

    def test_acceptance_simple(self):
        """
        Test a call routed to another number with call acceptance.

        Call Acceptance is a feature where the called channel first needs to
        complete a DTMF challenge in order to verify the call was picked up by
        a human (and not voicemail).
        """
        events_file = 'fixtures/mobile/acceptance_simple.json'
        events = self.run_and_get_events(events_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/voipgrid-siproute-docker-00000048',
                'targets': ['SIP/voipgrid-siproute-docker-00000049'],
            }),
            ('on_up', {
                'caller': 'SIP/voipgrid-siproute-docker-00000048',
                'target': 'SIP/voipgrid-siproute-docker-00000049',
            }),
            ('on_hangup', {
                'caller': 'SIP/voipgrid-siproute-docker-00000048',
                'reason': 'completed',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_acceptance_reject(self):
        """
        Test call acceptance where target does not accept call.
        """
        events_file = 'fixtures/mobile/acceptance_reject.json'
        events = self.run_and_get_events(events_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/voipgrid-siproute-docker-00000063',
                'targets': ['SIP/voipgrid-siproute-docker-00000064'],
            }),
            ('on_hangup', {
                'caller': 'SIP/voipgrid-siproute-docker-00000063',
                'reason': 'no-answer',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_acceptance_group_pickup(self):
        """
        Test call acceptance to group with call acceptance and timeout.

        In this example, both targets are picked up, but only one can complete
        call confirmation. The other target is hung up by Asterisk.
        """
        events_file = 'fixtures/mobile/acceptance_group_pickup.json'
        events = self.run_and_get_events(events_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/voipgrid-siproute-docker-00000069',
                'targets': [
                    'SIP/voipgrid-siproute-docker-0000006a',
                    'SIP/voipgrid-siproute-docker-0000006c',
                ],
            }),
            ('on_up', {
                'caller': 'SIP/voipgrid-siproute-docker-00000069',
                'target': 'SIP/voipgrid-siproute-docker-0000006c',
            }),
            ('on_hangup', {
                'caller': 'SIP/voipgrid-siproute-docker-00000069',
                'reason': 'completed',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_acceptance_group_reject(self):
        """
        Test call acceptance to a group with call acceptance and reject.
        """
        events_file = 'fixtures/mobile/acceptance_group_reject.json'
        events = self.run_and_get_events(events_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/voipgrid-siproute-docker-00000072',
                'targets': [
                    'SIP/voipgrid-siproute-docker-00000073',
                    'SIP/voipgrid-siproute-docker-00000074',
                ],
            }),
            ('on_up', {
                'caller': 'SIP/voipgrid-siproute-docker-00000072',
                'target': 'SIP/voipgrid-siproute-docker-00000073',
            }),
            ('on_hangup', {
                'caller': 'SIP/voipgrid-siproute-docker-00000072',
                'reason': 'completed',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_acceptance_timeout(self):
        """
        Test call acceptance to a group with call acceptance and reject.
        """
        events_file = 'fixtures/mobile/acceptance_timeout.json'
        events = self.run_and_get_events(events_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/voipgrid-siproute-docker-0000007b',
                'targets': ['SIP/voipgrid-siproute-docker-0000007c'],
            }),
            ('on_hangup', {
                'caller': 'SIP/voipgrid-siproute-docker-0000007b',
                'reason': 'no-answer',
            }),
        ]

        self.assertEqualChannels(expected_events, events)



from cacofonisk.callerid import CallerId
from cacofonisk.channel import SimpleChannel
from tests.replaytest import ChannelEventsTestCase


class TestSimple(ChannelEventsTestCase):

    def test_ab_success_a_hangup(self):
        """
        Test a simple, successful call where A hangs up.
        """
        fixture_file = 'fixtures/simple/ab_success_a_hangup.json'
        events = self.run_and_get_events(fixture_file)

        calling_chan = SimpleChannel(
            account_code='150010001',
            caller_id=CallerId(name='Andrew Garza', num='201'),
            cid_calling_pres='1 (Presentation Allowed, Passed Screen)',
            connected_line=CallerId(),
            exten='202',
            linkedid='195176c06ab8-1529936170.42',
            name='SIP/150010001-00000004',
            state=6,
            uniqueid='195176c06ab8-1529936170.42',
        )

        target_chan = SimpleChannel(
            account_code='150010001',
            caller_id=CallerId(num='202'),
            cid_calling_pres='0 (Presentation Allowed, Not Screened)',
            connected_line=CallerId(name='Andrew Garza', num='201'),
            exten='s',
            linkedid='195176c06ab8-1529936170.42',
            name='SIP/150010002-00000005',
            state=6,
            uniqueid='195176c06ab8-1529936170.50',
        )

        expected_events = [
            ('on_b_dial', {
                'caller': calling_chan.replace(state=4),
                'targets': [target_chan.replace(state=5)],
            }),
            ('on_up', {
                'caller': calling_chan,
                'target': target_chan,
            }),
            ('on_hangup', {
                'caller': calling_chan,
                'reason': 'completed',
            }),
        ]

        self.assertEqual(expected_events, events)

    def test_ab_success_b_hangup(self):
        """
        Test a simple, successful call where B hangs up.
        """
        fixture_file = 'fixtures/simple/ab_success_b_hangup.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/150010001-00000013',
                'targets': ['SIP/150010002-00000014'],
            }),
            ('on_up', {
                'caller': 'SIP/150010001-00000013',
                'target': 'SIP/150010002-00000014',
            }),
            ('on_hangup', {
                'caller': 'SIP/150010001-00000013',
                'reason': 'completed',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_success_twoclients(self):
        """
        Test a simple, successful call with calls through a proxy.

        Account 260010001 using +31260010001 as outdialing number, dials
        +31150010001 which is connected to account 150010001 with internal
        number 201 the call is picked up and completed successfully.
        """
        fixture_file = 'fixtures/simple/ab_success_twoclients.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/260010001-00000015',
                'targets': ['SIP/voipgrid-siproute-docker-00000016'],
            }),
            ('on_b_dial', {
                'caller': 'SIP/voipgrid-siproute-docker-00000017',
                'targets': ['SIP/150010001-00000018'],
            }),
            ('on_up', {
                'caller': 'SIP/voipgrid-siproute-docker-00000017',
                'target': 'SIP/150010001-00000018',
            }),
            ('on_up', {
                'caller': 'SIP/260010001-00000015',
                'target': 'SIP/voipgrid-siproute-docker-00000016',
            }),
            ('on_hangup', {
                'caller': 'SIP/voipgrid-siproute-docker-00000017',
                'reason': 'completed',
            }),
            ('on_hangup', {
                'caller': 'SIP/260010001-00000015',
                'reason': 'completed',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_reject(self):
        """
        Test a simple call where B rejects the call.
        """
        fixture_file = 'fixtures/simple/ab_reject.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/150010001-00000008',
                'targets': ['SIP/150010002-00000009'],
            }),
            ('on_hangup', {
                'caller': 'SIP/150010001-00000008',
                'reason': 'busy',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_dnd(self):
        """
        Test a simple call where B is set to Do Not Disturb.
        """
        fixture_file = 'fixtures/simple/ab_dnd.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_hangup', {
                'caller': 'SIP/150010001-00000006',
                'reason': 'busy',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_a_cancel_hangup(self):
        """
        Test a call where A hangs up before B can pick up.
        """
        fixture_file = 'fixtures/simple/ab_a_cancel.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/150010001-00000002',
                'targets': ['SIP/150010002-00000003'],
            }),
            ('on_hangup', {
                'caller': 'SIP/150010001-00000002',
                'reason': 'cancelled',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_callgroup(self):
        """
        Test a simple call to a group where one phone is picked up.
        """
        fixture_file = 'fixtures/simple/ab_callgroup.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/150010001-0000000d',
                'targets': [
                    'SIP/150010002-0000000e',
                    'SIP/150010003-0000000f',
                ],
            }),
            ('on_up', {
                'caller': 'SIP/150010001-0000000d',
                'target': 'SIP/150010002-0000000e',
            }),
            ('on_hangup', {
                'caller': 'SIP/150010001-0000000d',
                'reason': 'completed',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_callgroup_no_answer(self):
        """
        Test a call to a group where no target picks up.
        """
        fixture_file = 'fixtures/simple/ab_callgroup_no_answer.json'
        events = self.run_and_get_events(fixture_file)

        expected_events = [
            ('on_b_dial', {
                'caller': 'SIP/150010001-00000010',
                'targets': [
                    'SIP/150010002-00000012',
                    'SIP/150010003-00000011',
                ],
            }),
            ('on_hangup', {
                'caller': 'SIP/150010001-00000010',
                'reason': 'no-answer',
            }),
        ]

        self.assertEqualChannels(expected_events, events)

    def test_ab_out_of_order_dial(self):
        """
        Test a simple AB call where the state is changed before the dial.
        """
        fixture_file = 'fixtures/simple/ab_success_o3.json'
        events = self.run_and_get_events(fixture_file)

        calling_chan = SimpleChannel(
            name='SIP/100060001-0000000e',
            uniqueid='fc7a1e79d662-1539093894.250',
            linkedid='fc7a1e79d662-1539093894.250',
            account_code='100060001',
            caller_id=CallerId(name='Robert Hughes', num='201'),
            cid_calling_pres='1 (Presentation Allowed, Passed Screen)',
            connected_line=CallerId(),
            exten='202',
            state=6,
        )

        target_chan = SimpleChannel(
            name='SIP/100060002-0000000f',
            uniqueid='fc7a1e79d662-1539093894.258',
            linkedid='fc7a1e79d662-1539093894.250',
            account_code='100060001',
            caller_id=CallerId(num='202'),
            cid_calling_pres='0 (Presentation Allowed, Not Screened)',
            connected_line=CallerId(name='Robert Hughes', num='201'),
            exten='s',
            state=6,
        )

        expected_events = [
            ('on_b_dial', {
                'caller': calling_chan.replace(state=4),
                'targets': [target_chan.replace(state=5)],
            }),
            ('on_up', {
                'caller': calling_chan,
                'target': target_chan,
            }),
            ('on_hangup', {
                'caller': calling_chan,
                'reason': 'completed',
            }),
        ]

        self.assertEqual(expected_events, events)



from collections import namedtuple

from cacofonisk.callerid import CallerId


class MissingUniqueid(KeyError):
    pass


class Channel(object):
    """
    A Channel holds Asterisk channel state.

    It can be dialed, bridge, and updated with new data. All of the above is
    typical low level Asterisk channel behaviour.
    """

    def __init__(self, event):
        """
        Create a new channel instance.

        Args:
            event (dict): The attributes of a Newchannel event.
        """
        self.name = event['Channel']
        self.uniqueid = event['Uniqueid']
        self.linkedid = event['Linkedid']
        self.state = int(event['ChannelState'])
        self.exten = event['Exten']
        self.account_code = event['AccountCode']
        self.cid_calling_pres = None
        self.caller_id = CallerId(
            name=event['CallerIDName'],
            num=event['CallerIDNum'],
        )
        self.connected_line = CallerId(
            name=event['ConnectedLineName'],
            num=event['ConnectedLineNum'],
        )

        # Create vars which are used to store generated data based on other
        # events from Asterisk.
        self.fwd_local_bridge = None
        self.back_local_bridge = None
        self.back_dial = None
        self.fwd_dials = []
        self.bridge = None
        self.is_originated = False
        self.is_calling = False

        # The Custom dict can be used to store custom data.
        self.custom = {}

    def __repr__(self):
        return (
            '<Channel('
            'name={self.name!r} '
            'id={self.uniqueid!r} '
            'linkedid={self.linkedid!r} '
            'forward_local_bridge={next} '
            'backward_local_bridge={prev} '
            'state={self.state!r} '
            'cli="{self.caller_id.name!r}" <{self.caller_id.num!r}> '
            'exten={self.exten})>').format(
            self=self,
            next=(self.fwd_local_bridge and self.fwd_local_bridge.name),
            prev=(self.back_local_bridge and self.back_local_bridge.name))

    @property
    def is_local(self):
        """
        Whether the current channel is a local channel.

        A connection to Asterisk consist of two parts, an internal connection
        and an external connection. The internal connection is prefixed
        with 'Local/', whereas a external connection is prefixed with 'SIP/'.

        Returns:
            bool: True if the channel is local, false otherwise.
        """
        return self.name.startswith('Local/')

    @property
    def has_extension(self):
        """
        Check if the channel has a valid extension (not empty or star signal).

        Calling channels in Asterisk typically have an extension which
        tells Asterisk where it should direct the call. Checking whether the
        channel has an extension is usually a good indicator to determine
        whether the channel is calling or not.

        Returns:
            bool: Whether the channel has a valid extension.
        """
        return (
                self.exten and self.exten != 's' and
                not self.exten.startswith('*')
        )

    def get_dialing_channel(self):
        """
        Figure out on whose channel's behalf we're calling.

        When a channel is not bridged yet, you can use this on the
        B-channel to figure out which A-channel initiated the call.

        Returns:
            Channel: The master channel dialing this channel.
        """
        if self.back_dial:
            # Check if we are being dialed.
            a_chan = self.back_dial
        else:
            # This is the root channel.
            a_chan = None

        # If our a_chan has a local bridge, use the back part of that bridge
        # to check for further dials.
        if a_chan and a_chan.back_local_bridge:
            a_chan = a_chan.back_local_bridge

        # If we have an incoming channel, recurse through the channels to find
        # the true origin channel. If we don't have one, it means we're the
        # origin channel.
        return a_chan.get_dialing_channel() if a_chan else self

    def get_dialed_channels(self):
        """
        Figure out which channels are calling on our behalf.

        When a channel is not bridged yet, you can use this on the
        A-channel to find out which channels are dialed on behalf of
        this channel.

        It works like this:

        * A-channel (this) has a list of _fwd_dials items (open
          dials).
        * We loop over those (a list of uniqueids) and find the
          corresponding channels.
        * Those channels may be SIP channels, or they can be local
          channels, in which case we have to look further (by calling
          this function on those channels).

        Returns:
            set: A set of all channels being dialed by this channel.
        """
        b_channels = set()

        b_chans = self.fwd_dials

        for b_chan in b_chans:
            # Likely, b_chan.fwd_local_bridge is None, in which case we're
            # looking at a real tech channel (non-Local).
            # Or, the b_chan has one.fwd_local_bridge, after which we have
            # to call this function again.
            if b_chan.fwd_local_bridge:
                b_chan = b_chan.fwd_local_bridge

                assert not b_chan.fwd_local_bridge, (
                    'Since when does asterisk do double links? b_chan={!r}'
                    .format(b_chan)
                )

                b_channels.update(b_chan.get_dialed_channels())
            else:
                assert not b_chan.fwd_dials
                b_channels.add(b_chan)

        return b_channels

    def get_bridge_peers_recursive(self):
        """
        Traverse (local) bridges recursively and get all non-Local channels.

        Uses _get_peers_recursive to traverse the local and non-local
        bridges recursively, but also checks the local bridges of self for
        more channels.

        This cannot be done in one step, as doing so would cause a recursion
        loop while traversing the same local bridge back and forth.

        Returns:
            set: A set of non-local Channels.
        """
        peers = self._bridge_peers_recurse()

        if self.fwd_local_bridge:
            fwd_peers = self.fwd_local_bridge._bridge_peers_recurse()
            for peer in fwd_peers:
                peers.add(peer)

        if self.back_local_bridge:
            back_peers = self.back_local_bridge._bridge_peers_recurse()
            for peer in back_peers:
                peers.add(peer)

        return peers

    def _bridge_peers_recurse(self):
        """
        The recursive step of get_bridge_peers_recursive.

        This function finds the non-Local channels of the current bridge,
        as well as the non-Local channels of the channels linked through
        local bridges.

        Returns:
            set: A set of non-local Channels.
        """
        if not self.bridge:
            return set()

        peers = set()

        if not self.is_local:
            peers.add(self)

        for peer in [peer for peer in self.bridge.peers if peer != self]:
            if not peer.is_local:
                peers.add(peer)
            elif peer.is_local and peer.fwd_local_bridge:
                rec_peers = peer.fwd_local_bridge._bridge_peers_recurse()

                for recursive_peer in rec_peers:
                    peers.add(recursive_peer)
            elif peer.is_local and peer.back_local_bridge:
                rec_peers = peer.back_local_bridge._bridge_peers_recurse()

                for recursive_peer in rec_peers:
                    peers.add(recursive_peer)

        return peers

    def as_namedtuple(self):
        """
        Convert Channel to a SimpleChannel, so it's safe to pass to a reporter.

        Returns:
            SimpleChannel: A SimpleChannel with the data of this channel.
        """
        fields = SimpleChannel._fields

        return SimpleChannel(
            **{field: getattr(self, field) for field in fields})


class ChannelDict(dict):
    """
    ChannelDict is a dict which raises a MissingUniqueid if the key is missing.
    """

    def __getitem__(self, item):
        try:
            return super(ChannelDict, self).__getitem__(item)
        except KeyError:
            raise MissingUniqueid(item)


class SimpleChannel(namedtuple(
    'SimpleChannelBase', 'name uniqueid linkedid account_code caller_id '
                         'cid_calling_pres connected_line exten state'
)):
    def replace(self, **kwargs):
        """
        Make the _replace method public.

        Args:
            **kwargs: The fields to replace.

        Returns:
            A new instance of SimpleChannel.
        """
        return self._replace(**kwargs)
